use std::collections::HashSet;
use std::fs::File;
use std::hash::Hash;
use std::path::PathBuf;
use criterion::{black_box, criterion_group, criterion_main, Criterion};
use fddiscovery::edgecollectors::EdgeConsumer;
use fddiscovery::validation::plicache::{PLICache, PLICacheOps};
use hittingset::{DefaultEdge, Edge};
use plistore::{Cluster, ClusterFilter, CSVConfig, PLITable, read_input};
use serde::{Deserialize, Serialize};
use fddiscovery::problems::fds::FDProblem;
use fddiscovery::sampling::SimpleSampling;

pub struct EdgeHashCollector<E: Edge + Eq + Hash>(HashSet<E>);

impl<E: Edge + Eq + Hash> EdgeHashCollector<E> {
    pub fn new() -> EdgeHashCollector<E> {
        EdgeHashCollector(HashSet::new())
    }

    pub fn get_edge_set(&self) -> &HashSet<E> {
        &self.0
    }
}

impl<E: Edge + Eq + Hash> EdgeConsumer<E> for EdgeHashCollector<E> {
    fn consume(&mut self, edge: E) {
        self.0.insert(edge);
    }
}

impl ClusterFilter for EdgeHashCollector<DefaultEdge> {
    fn keep(&mut self, cluster: &Cluster) -> bool {
        cluster.len() > 1
    }

    fn set_requested_level(&mut self, active: bool) {
    }
}

fn bench(table: &PLITable, test: &Vec<PLICacheOps<DefaultEdge>>) -> usize {
    let mut filter =  EdgeHashCollector::new();
    let iter = test.iter();
    let problem = FDProblem::new(0, table);
    let sampling = SimpleSampling::new(0.3, &problem);
    let mut cache = PLICache::new(table, &sampling);
    for e in iter {
        match e {
            PLICacheOps::GET_PLI(columns, cand) => { cache.get_pli(columns, &mut filter, cand); },
            PLICacheOps::TRUNCATE(level) => cache.truncate(*level),
        }
    }
    filter.get_edge_set().len()
}


#[derive(Default, Debug, Deserialize, Serialize)]
pub struct InputConfig {
    files: Vec<PathBuf>,
    config: CSVConfig,
}
fn criterion_benchmark(c: &mut Criterion) {
    let paths = glob::glob("../profiles/pli/*.json").unwrap();


    for path in paths {
        let config: InputConfig = serde_yaml::from_reader(File::open("../data/flight.yaml").unwrap()).unwrap();
        let plitable = read_input(&config.config, &PathBuf::from("../data/flight_r1001_c109.csv")).unwrap();
        let input : Vec<PLICacheOps<DefaultEdge>> = serde_json::Deserializer::from_reader(File::open(path.as_ref().unwrap().as_path()).unwrap()).into_iter::<PLICacheOps<DefaultEdge>>().filter_map(|x| x.ok()).collect();
        c.bench_function(path.as_ref().unwrap().file_name().unwrap().to_str().unwrap(), |b| b.iter(|| bench(black_box(&plitable),black_box(&input))));
    }


}

criterion_group!(benches, criterion_benchmark);
criterion_main!(benches);