use std::error::Error;
use std::fs::File;
use std::path::{Path, PathBuf};
use std::{fs, process};
use std::io::{LineWriter, Write};
use std::time::Instant;
use clap::{Parser, Subcommand};
use serde::{Serialize, Deserialize};
use serde_json::json;
use shadow_rs::shadow;
use strum_macros::AsRefStr;
use tracing::Dispatch;
use tracing_timing::{Builder, Histogram, TimingSubscriber};

use fddiscovery::edgecollectors::{EdgeConsumer, ResultCounter, VecEdgeCollector};
use plistore::{CSVConfig, PLITable, read_input};
use crate::OptionalResultCollector::{Collector, Counter};

#[derive(Parser)]
#[clap(author, version, about, long_about = None)]
struct Cli {
    #[clap(subcommand)]
    command: Commands,

    /// Sets a custom config file
    #[clap(parse(from_os_str), value_name = "INPUT")]
    input: PathBuf,

    /// Sets a custom config file
    #[clap(parse(from_os_str), value_name = "OUTPUT")]
    output: PathBuf,

    #[clap(short, long)]
    enumerate_results: bool,

    #[clap(short, long, default_value = "1")]
    iterations: u32,

    #[clap(short, long)]
    profile: bool,

    #[clap(short, long, default_value = "0.3")]
    sampling_factor: f64

}

#[derive(Subcommand, AsRefStr)]
enum Commands {
    FD {
        #[clap(short, long)]
        sort: bool,
    },
    FDPARALLEL {
    },
    FDUNION {},
}

#[derive(Default, Debug, Deserialize, Serialize)]
pub struct InputConfig {
    files: Vec<PathBuf>,
    config: CSVConfig,
}

pub enum OptionalResultCollector<E> {
    Counter(ResultCounter),
    Collector(VecEdgeCollector<E>)
}

impl<E> OptionalResultCollector<E> {
    fn result_count(&self) -> usize {
        match self {
            Counter(c) => c.count(),
            Collector(c) => c.results.len()
        }
    }

    fn results(&self) -> Option<&Vec<E>> {
        match self {
            Counter(c) => None,
            Collector(c) => Some(&c.results)
        }
    }
}

impl<E> EdgeConsumer<E> for OptionalResultCollector<E> {
    fn consume(&mut self, edge: E) {
        match self {
            Counter(c) => c.consume(edge),
            Collector(c) => c.consume(edge)
        }
    }
}

fn run() -> Result<(), Box<dyn Error>> {
    shadow!(build);

    let cli = Cli::parse();

    let subscriber = Builder::default().no_span_recursion().span_close_events().build(|| Histogram::new_with_max(1_000_000_000, 3).unwrap());
    let dispatch = Dispatch::new(subscriber);

    if cli.profile {
        tracing::dispatcher::set_global_default(dispatch.clone())
            .expect("setting tracing default failed");
    }

    let output_path = Path::new(&cli.output);
    let file = File::create(output_path)?;
    let mut output = LineWriter::new(file);


    let config: InputConfig = serde_yaml::from_reader(File::open(&cli.input)?)?;

    for input_path in &config.files {
        for _i in 0..cli.iterations {
            let start = Instant::now();
            let preprocessing_start = Instant::now();
            let mut plis = read_input(&config.config, &cli.input.parent().unwrap().join(input_path))?;
            let preprocessing_end = Instant::now();
            println!("Preprocessing: {:?}", preprocessing_end - preprocessing_start);
            let mut collector1 = if cli.enumerate_results { Collector(VecEdgeCollector::new()) } else { Counter(ResultCounter::new())};
            match &cli.command {
                Commands::FD { sort } => {
                    fddiscovery::problems::fds::search_fds(&mut plis, *sort, cli.sampling_factor,&mut collector1);
                }
                Commands::FDPARALLEL {} => {
                    fddiscovery::problems::fds::search_fds_parallel(&mut plis, cli.sampling_factor, &mut collector1);
                }
                Commands::FDUNION {} => {
                    fddiscovery::problems::fdsunion::search_fds(&plis, cli.sampling_factor, &mut collector1);
                }
            }
            if cli.profile {
                print_stats(&dispatch);
            }
            let time =  Instant::now() - start;
            let count = collector1.result_count();
            let mut result = json!({
                "sampling_factor": cli.sampling_factor,
                "runtime": time.as_secs_f64(),
                "results": collector1.results(),
                "result_size": count,
                "problem": cli.command.as_ref(),
                "input": input_path,
                "config": &config.config,
                "rows": plis.row_count(),
                "columns": plis.column_count(),
                "build": {
                    "commit": build::COMMIT_HASH,
                    "debug": shadow_rs::is_debug(),
                    "branch": shadow_rs::branch(),
                    "git-clean": shadow_rs::git_clean(),
                    "git-status": shadow_rs::git_status_file(),
                    "build-time": build::BUILD_TIME_3339,
                },
            });
            output.write_all(result.to_string().as_bytes())?;
            writeln!(output)?;

            println!("Result size: {} Time: {:?}", count, time);
        }
    }





    Ok(())
}

fn print_stats(dispatch: &Dispatch) {
    dispatch.downcast_ref::<TimingSubscriber>().unwrap().force_synchronize();
    dispatch.downcast_ref::<TimingSubscriber>().unwrap().with_histograms(|hs| {
        for (span_group, hs) in hs {
            for (event_group, h) in hs {
                println!(
                    "{} -> {}: mean: {:.1}µs, p50: {}µs, p90: {}µs, p99: {}µs, p999: {}µs, max: {}µs, total: {}µs, count: {}",
                    span_group,
                    event_group,
                    h.mean() / 1000.0,
                    h.value_at_quantile(0.5) / 1_000,
                    h.value_at_quantile(0.9) / 1_000,
                    h.value_at_quantile(0.99) / 1_000,
                    h.value_at_quantile(0.999) / 1_000,
                    h.max() / 1_000,
                    h.mean() * h.len() as f64 / 1000.0,
                    h.len()
                );
                // for v in break_once(
                //     h.iter_linear(500).skip_while(|v| v.quantile() < 0.01),
                //     |v| v.quantile() > 0.90,
                // ) {
                //     println!(
                //         "{:4.1}µs | {:40} | {:4.1}th %-ile",
                //         (v.value_iterated_to() + 1) as f64 / 1_000.0,
                //         "*".repeat(
                //             (v.count_since_last_iteration() as f64 * 40.0 / h.len() as f64).ceil() as usize
                //         ),
                //         v.percentile(),
                //     );
                // }
                h.reset();
            }
        }
    });
}


fn main() {
    if let Err(err) = run() {
        println!("{}", err);
        process::exit(1);
    }
}

// until we have https://github.com/rust-lang/rust/issues/62208
fn break_once<I, F>(it: I, mut f: F) -> impl Iterator<Item = I::Item>
    where
        I: IntoIterator,
        F: FnMut(&I::Item) -> bool,
{
    let mut got_true = false;
    it.into_iter().take_while(move |i| {
        if got_true {
            // we've already yielded when f was true
            return false;
        }
        if f(i) {
            // this must be the first time f returns true
            // we should yield i, and then no more
            got_true = true;
        }
        // f returned false, so we should keep yielding
        true
    })
}
