use std::collections::HashSet;
use std::hash::Hash;
use std::sync::mpsc;
use std::sync::mpsc::{Receiver, Sender};

use itertools::Itertools;

use hittingset::{Edge, Hypergraph, ImplicableEdge};

pub trait EdgeConsumer<E> {
    fn consume(&mut self, edge: E);
}

pub struct EdgeHashCollector<E: Edge + Eq + Hash>(HashSet<E>);

impl<E: Edge + Eq + Hash + ImplicableEdge> EdgeHashCollector<E> {
    pub fn new() -> EdgeHashCollector<E> {
        EdgeHashCollector(HashSet::new())
    }

    pub fn get_edge_set(&self) -> &HashSet<E> {
        &self.0
    }

    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    pub fn to_graph(&mut self, candidate: &E) -> Hypergraph<E> {
        let mut result = Hypergraph::new(candidate.len());
        // for e in self.0.drain() {
        //     result.add_edge_and_minimize(e);
        // }
        self.0.drain().sorted_unstable().for_each(|e| {
            if !result.implies(&e) {
                result.add_edge(e);
            }
        });

        result
    }
}

impl<E: Edge + Eq + Hash> EdgeConsumer<E> for EdgeHashCollector<E> {
    fn consume(&mut self, edge: E) {
        self.0.insert(edge);
    }
}

pub struct ResultCounter {
    count: usize,
}

impl<T> EdgeConsumer<T> for ResultCounter {
    fn consume(&mut self, _: T) {
        self.count += 1;
    }
}

impl ResultCounter {
    pub fn new() -> ResultCounter { ResultCounter { count: 0 } }

    pub fn count(&self) -> usize { self.count }
}

pub struct VecEdgeCollector<E> {
    pub results: Vec<E>,
}

impl<E> VecEdgeCollector<E> {
    pub fn new() -> VecEdgeCollector<E> {
        VecEdgeCollector {
            results: vec![]
        }
    }
}

impl<E> EdgeConsumer<E> for VecEdgeCollector<E> {
    fn consume(&mut self, a: E) {
        self.results.push(a);
    }
}

#[derive(Clone)]
pub struct ChannelEdgeCollector<E> {
    tx: Sender<E>,
}

impl<E> ChannelEdgeCollector<E> {
    pub fn new() -> (ChannelEdgeCollector<E>, Receiver<E>) {
        let (tx, rx): (Sender<E>, Receiver<E>) = mpsc::channel();
        (ChannelEdgeCollector {
            tx
        }, rx)
    }
}

impl<E> EdgeConsumer<E> for ChannelEdgeCollector<E> {
    fn consume(&mut self, edge: E) {
        self.tx.send(edge).unwrap();
    }
}
