use std::borrow::BorrowMut;
use std::cmp::max;
use std::collections::HashSet;
use std::fs::File;

use rayon::prelude::*;
use tracing::{trace_span};

use hittingset::{DefaultEdge, Edge, Hypergraph, search};
#[cfg(feature = "collect_profiles")]
use hittingset::PROFILE_FILE;
use plistore::{PLITable, RowIndex};

use crate::edgecollectors::{ChannelEdgeCollector, EdgeConsumer, EdgeHashCollector};
use crate::problems::common::get_default_edge;
use crate::problems::ProblemConfig;
use crate::problems::results::MultiFD;
use crate::problems::uccs::UCCProblem;
use crate::sampling::{SamplingStrategy, SimpleSampling, SortedSampling};
use crate::validation::PLIValidator;

pub struct FDProblem<'a> {
    columnid: usize,
    table: &'a PLITable,
}

impl<'a> FDProblem<'a> {
    pub fn new(columnid: usize, table: &'a PLITable) -> Self {
        Self { columnid, table }
    }

    fn is_valid_edge(&self, edge: &DefaultEdge) -> bool {
        edge.contains(self.columnid)
    }
}

impl<'a> ProblemConfig for FDProblem<'a> {
    type EdgeType = DefaultEdge;

    fn get_edge(&self, p1: RowIndex, p2: RowIndex, receiver: &mut impl EdgeConsumer<DefaultEdge>) {
        let edge = get_default_edge(self.table, p1, p2);
        if self.is_valid_edge(&edge) {
            receiver.consume(edge)
        }
    }

    fn is_valid_result(&self, result: &DefaultEdge) -> bool {
        !result.contains(self.columnid)
    }

    fn is_violating_pair(&self, r1: RowIndex, r2: RowIndex, _candidate: &Self::EdgeType) -> bool {
        self.table.is_different_on_column(r1, r2, self.columnid)
    }
}

struct FDTranslator<'a, N> {
    target: usize,
    next: &'a mut N,
}

impl<'a, N: EdgeConsumer<MultiFD>> EdgeConsumer<DefaultEdge> for FDTranslator<'a, N> {
    fn consume(&mut self, result: DefaultEdge) {
        let mut rhs = DefaultEdge::empty(result.len());
        rhs.add(self.target);
        self.next.consume(MultiFD {
            lhs: result,
            rhs,
        })
    }
}

pub fn search_fds(table: &mut PLITable, use_sorting: bool, sampling_factor: f64, consumer: &mut impl EdgeConsumer<MultiFD>) {
    let mut sample = HashSet::new();
    for target in 0..table.column_count() {
        println!("Target {}", target);
        if use_sorting {
            trace_span!("sort").in_scope(|| {
                table.sort_plis_by(target);
            });
        }

        let problem = FDProblem { columnid: target, table };

        if use_sorting {
            let sampling = SortedSampling::new(target, table, sampling_factor, &problem);
            search_fds_for_target(table, consumer, &mut sample, &problem, sampling);
        } else {
            let sampling = SimpleSampling::new(sampling_factor, &problem);
            search_fds_for_target(table, consumer, &mut sample, &problem, sampling);
        }


    }
}

fn search_fds_for_target<'a>(
    table: &'a PLITable,
    consumer: &mut impl EdgeConsumer<MultiFD>,
    sample: &mut HashSet<DefaultEdge>,
    problem: &FDProblem,
    sampling: impl SamplingStrategy<DefaultEdge>
) {

    let mut input = Hypergraph::new(table.column_count());
    let mut collector = EdgeHashCollector::new();
    for e in sample.iter() {
        if !problem.is_valid_edge(e) { continue; }
        input.add_edge_and_minimize(e.clone());
    }
    for c in 0..table.column_count() {
        if c != problem.columnid {
            sampling.sample_pli(table.get_column_pli(c), &mut collector);
        }
    }

    for e in collector.get_edge_set() {
        if !problem.is_valid_edge(e) { continue; }
        input.add_edge_and_minimize(e.clone());
    }

    let mut translator = FDTranslator { target: problem.columnid, next: consumer };

    #[cfg(feature = "collect_profiles")]
    PROFILE_FILE.with(|f| *f.borrow_mut() = Some(std::fs::OpenOptions::new().create(true).write(true).open(format!("profiles/edgemarkstore-{}-{}.json", table.get_name().as_deref().unwrap_or("Unknown"), problem.columnid)).unwrap()));
    #[cfg(feature = "collect_pli_profiles")]
    crate::validation::plicache::PROFILE_FILE.with(|f| *f.borrow_mut() = Some(std::fs::OpenOptions::new().create(true).truncate(true).write(true).open(format!("profiles/pli/pli-{}-{}.json", table.get_name().as_deref().unwrap_or("Unknown"), problem.columnid)).unwrap()));


    let mut validator = PLIValidator::new(table, &mut translator, problem, &sampling);
    search(&mut input, &mut validator);

    #[cfg(feature = "collect_profiles")]
    PROFILE_FILE.with(|f| {
        *f.borrow_mut() = None;
    });
    #[cfg(feature = "collect_pli_profiles")]
    crate::validation::plicache::PROFILE_FILE.with(|f| {
        *f.borrow_mut() = None;
    });
    for s in input {
        sample.insert(s);
    }
}

pub fn search_fds_parallel(table: &mut PLITable, sampling_factor: f64, consumer: &mut impl EdgeConsumer<MultiFD>) {
    let (collector, rx) = ChannelEdgeCollector::new();
    (0..table.column_count()).into_par_iter().for_each_with(collector, |collector, target| {
        let problem = UCCProblem::new(table);
        let sampling = SimpleSampling::new(sampling_factor, &problem);
        let mut collector = collector.clone();
        sampling.sample_pli(table.get_column_pli(target), &mut collector);
    });
    let mut sample = HashSet::new();
    while let Ok(result) = rx.recv() {
        sample.insert(result);
    }

    let (collector, rx) = ChannelEdgeCollector::new();
    (0..table.column_count()).into_par_iter().for_each_with(collector, |consumer, target| {
        let mut input = Hypergraph::new(table.column_count());
        let problem = FDProblem { columnid: target, table };
        let sampling = SimpleSampling::new(sampling_factor, &problem);
        for e in sample.iter() {
            if !problem.is_valid_edge(e) { continue; }
            input.add_edge_and_minimize(e.clone());
        }

        let mut translator = FDTranslator { target, next: consumer };

        let mut validator = PLIValidator::new(table, &mut translator, &problem, &sampling);
        search(&mut input, &mut validator);
    });
    while let Ok(result) = rx.recv() {
        consumer.consume(result);
    }
}