use std::time::Instant;

use hittingset::{DefaultEdge, Edge, Hypergraph, ImplicableEdge};
use plistore::{PLITable, RowIndex};

use crate::edgecollectors::{EdgeConsumer, EdgeHashCollector};
use crate::problems::ProblemConfig;
use crate::problems::results::MultiFD;
use crate::sampling::{SamplingStrategy, SimpleSampling};
use crate::treesearch::treesearchfd::search;
use crate::validation::PLIValidator;

struct FDProblemUnion<'a> {
    table: &'a PLITable,
    separate_edges: bool,
}

impl<'a> ProblemConfig for FDProblemUnion<'a> {
    type EdgeType = MultiFD;

    fn get_edge(&self, p1: RowIndex, p2: RowIndex, receiver: &mut impl EdgeConsumer<MultiFD>) {
        if self.separate_edges {
            let mut lhs = DefaultEdge::empty(self.table.column_count());
            for column in 0..self.table.column_count() {
                if self.table.is_different_on_column(p1, p2, column) {
                    lhs.add(column);
                }
            }
            for column in lhs.fast_nodes() {
                let mut rhs = DefaultEdge::filled(self.table.column_count());
                rhs.remove(column);
                receiver.consume(MultiFD {
                    lhs: lhs.clone(),
                    rhs,
                });
            }
        } else {
            let mut edge = MultiFD {
                lhs: DefaultEdge::empty(self.table.column_count()),
                rhs: DefaultEdge::empty(self.table.column_count()),
            };
            for column in 0..self.table.column_count() {
                if self.table.is_different_on_column(p1, p2, column) {
                    edge.lhs.add(column);
                } else {
                    edge.rhs.add(column);
                }
            }
            receiver.consume(edge)
        }
    }

    fn is_valid_result(&self, result: &MultiFD) -> bool {
        true
    }

    fn is_violating_pair(&self, r1: RowIndex, r2: RowIndex, candidate: &Self::EdgeType) -> bool {
        candidate.rhs.fast_nodes().any(|c| self.table.is_different_on_column(r1, r2, c))
    }
}

pub fn search_fds(table: &PLITable, sampling_factor: f64, consumer: &mut impl EdgeConsumer<MultiFD>) {
    let mut consumer = MultiFDToSingleFD {
        next: consumer
    };
    let problem = FDProblemUnion { table, separate_edges: false };
    let sampling = SimpleSampling::new(sampling_factor, &problem);
    let mut collector = EdgeHashCollector::new();
    for c in 0..table.column_count() {
        sampling.sample_pli(table.get_column_pli(c), &mut collector);
    }
    let mut input = collector.to_graph(&MultiFD::empty(table.column_count() * 2));

    // this only makes sense with separate edges (otherwise it is a no-op)
    if problem.separate_edges {
        input = minimize(input);
    }


    let start = Instant::now();
    let mut validator = PLIValidator::new(table, &mut consumer, &problem, &sampling);
    search(&mut input, &mut validator);
    println!("Search {:?}", Instant::now() - start);

    println!("Edge count: {}", input.edge_count());
}

struct MultiFDToSingleFD<'a, N> {
    next: &'a mut N,
}

impl<'a, N: EdgeConsumer<MultiFD>> EdgeConsumer<MultiFD> for MultiFDToSingleFD<'a, N> {
    fn consume(&mut self, result: MultiFD) {
        for rhs in result.rhs.fast_nodes() {
            let mut b = DefaultEdge::empty(result.lhs.len());
            b.add(rhs);
            self.next.consume(MultiFD {
                lhs: result.lhs.clone(),
                rhs: b,
            })
        }

    }
}

fn minimize(raw_input: Hypergraph<MultiFD>) -> Hypergraph<MultiFD> {
    let mut input = Hypergraph::new(raw_input.vertex_count());

    println!("Start minimizing {}", raw_input.edge_count());
    for e in &raw_input {
        let mut rhs = e.rhs.clone();
        for e2 in &raw_input {
            if e2.lhs.implies(&e.lhs) {
                rhs.intersect_with(&e2.rhs);
            }
        }
        input.add_edge_and_minimize(MultiFD {
            lhs: e.lhs.clone(),
            rhs,
        })
    }
    println!("Finish minimizing {}", input.edge_count());
    input
}
