use std::fmt;
use std::fmt::Formatter;
use itertools::Itertools;
use serde::{Serialize};
use hittingset::{DefaultEdge, Edge, ImplicableEdge};

#[derive(Debug, PartialEq, Eq, Hash, Clone, Ord, PartialOrd, Serialize)]
pub struct MultiFD {
    pub lhs: DefaultEdge,
    pub rhs: DefaultEdge,
}

impl Edge for MultiFD {
    fn filled(size: usize) -> Self {
        MultiFD {
            lhs: DefaultEdge::filled(size / 2),
            rhs: DefaultEdge::filled(size / 2),
        }
    }

    fn empty(size: usize) -> Self {
        MultiFD {
            lhs: DefaultEdge::empty(size / 2),
            rhs: DefaultEdge::empty(size / 2),
        }
    }

    fn nodes(&'_ self) -> Box<dyn Iterator<Item=usize> + '_> {
        Box::new(self.lhs.fast_nodes().chain(self.rhs.fast_nodes().map(|s| s + self.lhs.len())))
    }


    fn add(&mut self, node: usize) {
        if node < self.lhs.len() {
            self.lhs.add(node);
        } else {
            self.rhs.add(node - self.lhs.len());
        }
    }

    fn remove(&mut self, node: usize) {
        if node < self.lhs.len() {
            self.lhs.remove(node);
        } else {
            self.rhs.remove(node - self.lhs.len());
        }
    }

    fn overlap(&self, other: &Self) -> usize {
        self.lhs.overlap(&other.lhs) + self.rhs.overlap(&other.rhs)
    }

    fn covers(&self, other: &Self) -> bool {
        self.lhs.covers(&other.lhs) || self.rhs.implies(&other.rhs)
    }

    fn remove_all(&mut self, other: &Self) {
        self.lhs.remove_all(&other.lhs);
        self.rhs.remove_all(&other.rhs);
    }

    fn add_all(&mut self, other: &Self) {
        self.lhs.add_all(&other.lhs);
        self.rhs.add_all(&other.rhs);
    }

    fn intersect(&self, other: &Self) -> Self {
        MultiFD {
            lhs: self.lhs.intersect(&other.lhs),
            rhs: self.rhs.intersect(&other.rhs),
        }
    }

    fn remove_all_invalid(&mut self, other: &Self) {
        for rhs in other.rhs.fast_nodes() {
            self.lhs.remove(rhs);
        }
        if other.rhs.fast_nodes().next().is_some() {
            self.rhs.clear();
        } else {
            for lhs in other.lhs.fast_nodes() {
                self.rhs.remove(lhs);
            }
        };
    }

    fn set_to(&mut self, other: &Self) {
        self.lhs.set_to(&other.lhs);
        self.rhs.set_to(&other.rhs);
    }

    fn contains(&self, node: usize) -> bool {
        if node < self.lhs.len() {
            self.lhs.contains(node)
        } else {
            self.rhs.contains(node - self.lhs.len())
        }
    }

    fn vertex_count(&self) -> usize {
        self.lhs.vertex_count() + self.rhs.vertex_count()
    }

    fn len(&self) -> usize {
        self.lhs.len() + self.rhs.len()
    }
}

impl MultiFD {

    pub fn fast_nodes(&'_ self) -> impl Iterator<Item=usize> + '_ {
        self.lhs.fast_nodes().chain(self.rhs.fast_nodes().map(|s| s + self.lhs.len()))
    }
}

impl fmt::Display for MultiFD {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "{} -> {}", self.lhs.fast_nodes().join(","), self.rhs.fast_nodes().join(","))
    }
}

impl ImplicableEdge for MultiFD {
    fn implies(&self, other: &Self) -> bool {
        self.lhs.implies(&other.lhs) && self.rhs.implies(&other.rhs)
    }

    fn is_implied_by(&self, others: &[Self]) -> bool {
        others.iter().any(|other| other.implies(self))
    }
}
