use std::cmp::max;
use itertools::{Itertools};

use rand::distributions::{Distribution, WeightedIndex};
use rand::prelude::ThreadRng;
use rand::Rng;
use hittingset::{DefaultEdge, Edge};

use plistore::{Cluster, ClusterIndex, PLITable, RowIndex};
use crate::edgecollectors::EdgeConsumer;
use crate::problems::fds::FDProblem;
use crate::problems::ProblemConfig;

pub trait SamplingStrategy<E : Edge> {
    fn sample_pli(&self, pli: &[Cluster], receiver: &mut impl EdgeConsumer<E>);
    fn sample(&self, pli: &[Cluster]) -> Vec<(ClusterIndex, RowIndex, RowIndex)>;
    fn get_cluster_size(&self, cluster: &Cluster) -> u64;
    fn get_cross_product(&self, cluster: &Cluster, consumer: &mut impl EdgeConsumer<E>);
    fn get_violating_pair(&self, cluster: &Cluster, candidate: &E) -> Option<(RowIndex, RowIndex)>;
}

pub struct SimpleSampling<'a, P: ProblemConfig> {
    sampling_factor: f64,
    problem: &'a P,
}

impl<'a, P: ProblemConfig> SimpleSampling<'a, P> {
    pub fn new(sampling_factor: f64, problem: &'a P) -> Self {
        Self { sampling_factor, problem }
    }
}

impl<'a, P: ProblemConfig> SamplingStrategy<P::EdgeType> for SimpleSampling<'a, P> {
    fn sample_pli(&self, pli: &[Cluster], receiver: &mut impl EdgeConsumer<P::EdgeType>) {
        if pli.is_empty() { return; }
        for (c_idx, r1, r2) in self.sample(pli) {
            self.problem.get_edge(pli[c_idx as usize][r1 as usize], pli[c_idx as usize][r2 as usize], receiver);
        }
    }

    fn sample(&self, pli: &[Cluster]) -> Vec<(ClusterIndex, RowIndex, RowIndex)> {
        let mut weights: Vec<u64> = Vec::new();
        let mut total_pairs: u64 = 0;
        for cluster in pli {
            let pairs = (cluster.len() * (cluster.len() - 1) / 2) as u64;
            total_pairs += pairs;
            weights.push(pairs);
        }

        let to_view = max(1, (total_pairs as f64).powf(self.sampling_factor) as usize);

        let dist = WeightedIndex::new(weights).unwrap();
        let mut rng = rand::thread_rng();
        let mut samples: Vec<(ClusterIndex, RowIndex, RowIndex)> = Vec::new();
        for _ in 0..to_view {
            let cluster = dist.sample(&mut rng);
            let size = pli[cluster].len() as RowIndex;
            let r1 = rng.gen_range(0..size as RowIndex);
            let r2 = (r1 + rng.gen_range(1..size as RowIndex)) % size;
            samples.push((cluster as ClusterIndex, r1, r2));
        }
        samples.sort_unstable();
        samples.dedup();
        samples
    }

    fn get_cluster_size(&self, cluster: &Cluster) -> u64 {
        (cluster.len() * (cluster.len() - 1) / 2) as u64
    }

    fn get_cross_product(&self, cluster: &Cluster, consumer: &mut impl EdgeConsumer<P::EdgeType>) {
        cluster.iter().copied().tuple_combinations().for_each(| (el1, el2) | self.problem.get_edge(el1, el2, consumer) );
    }

    fn get_violating_pair(&self, cluster: &Cluster, candidate: &P::EdgeType) -> Option<(RowIndex, RowIndex)> {
        let (r1, rest) = cluster.split_first().unwrap();
        rest.iter().find(|&r2| self.problem.is_violating_pair(*r1, *r2, candidate)).map(|r2| (*r1, *r2))
    }
}

pub struct SortedSampling<'a> {
    columnid: usize,
    table: &'a PLITable,
    sampling_factor: f64,
    problem: &'a FDProblem<'a>,
}

impl<'a> SamplingStrategy<DefaultEdge> for SortedSampling<'a> {
    fn sample_pli(&self, pli: &[Cluster], receiver: &mut impl EdgeConsumer<DefaultEdge>) {
        if pli.is_empty() { return; }
        for (c_idx, r1, r2) in self.sample(pli) {
            self.problem.get_edge(pli[c_idx as usize][r1 as usize], pli[c_idx as usize][r2 as usize], receiver);
        }
    }


    fn sample(&self, pli: &[Cluster]) -> Vec<(ClusterIndex, RowIndex, RowIndex)> {
        let mut weights: Vec<u64> = Vec::new();
        let mut total_pairs: u64 = 0;
        for cluster in pli {
            let pairs = self.get_cluster_size(cluster);
            total_pairs += pairs;
            weights.push(pairs);
        }
        if total_pairs == 0 { return vec![]; }

        let to_view = max(1, (total_pairs as f64).powf(self.sampling_factor) as usize);

        let dist = WeightedIndex::new(weights).unwrap();
        let mut rng = rand::thread_rng();
        let mut samples: Vec<(ClusterIndex, RowIndex, RowIndex)> = Vec::new();
        for _ in 0..to_view {
            let cluster = dist.sample(&mut rng);
            let cluster_pli = &pli[cluster];
            let (r1, r2) = self.sample_inter_cluster_positions(&mut rng, cluster_pli);
            samples.push((cluster as ClusterIndex, r1, r2));
        }
        samples.sort_unstable();
        samples.dedup();
        samples
    }

    fn get_cluster_size(&self, cluster: &Cluster) -> u64 {
        let mut iter = cluster.windows(2).enumerate();
        let mut start = 0;
        let mut len = 0;
        while let Some((pos, [l, r])) = iter.next() {
            if self.table.is_different_on_column(*l, *r, self.columnid) {
                len += (pos + 1 - start) * (cluster.len() - (pos + 1));
                start = pos + 1;
            }
        }
        len as u64
    }

    fn get_cross_product(&self, cluster: &Cluster, consumer: &mut impl EdgeConsumer<DefaultEdge>) {
        let mut iter = cluster.windows(2).enumerate();
        let mut start = 0;
        while let Some((pos, [l, r])) = iter.next() {
            if self.table.is_different_on_column(*l, *r, self.columnid) {
                for r1 in start..=pos {
                    for r2 in pos + 1..cluster.len() {
                        self.problem.get_edge(cluster[r1], cluster[r2], consumer);
                    }
                }
                start = pos + 1;
            }
        }
    }

    fn get_violating_pair(&self, cluster: &Cluster, candidate: &DefaultEdge) -> Option<(RowIndex, RowIndex)> {
        let r1 = *cluster.first().unwrap();
        let r2 = *cluster.last().unwrap();
        if self.problem.is_violating_pair(r1, r2, candidate) {
            Some((r1, r2))
        } else {
            None
        }
    }
}

impl<'a> SortedSampling<'a> {

    pub(crate) fn new(columnid: usize, table: &'a PLITable, sampling_factor: f64, problem: &'a FDProblem<'a>) -> Self {
        Self { columnid, table, sampling_factor, problem }
    }

    fn sample_inter_cluster_positions(&self, rng: &mut ThreadRng, cluster_pli: &Cluster) -> (RowIndex, RowIndex) {
        let mut upper_bound = cluster_pli.len() - 2;
        let last_tuple_id = cluster_pli[cluster_pli.len() - 1];
        while !self.table.is_different_on_column(cluster_pli[upper_bound], last_tuple_id, self.columnid) {
            upper_bound -= 1;
        }
        let r1 = rng.gen_range(0..=upper_bound as RowIndex);
        let mut lower_bound = r1 + 1;
        while !self.table.is_different_on_column(cluster_pli[lower_bound as usize], cluster_pli[r1 as usize], self.columnid) {
            lower_bound += 1;
        }
        let r2 = rng.gen_range(lower_bound..cluster_pli.len() as RowIndex);
        (r1, r2)
    }
}
