use fixedbitset::FixedBitSet;

use hittingset::bitutil::{intersect_with_inverse, is_subset_inverse, set_to, set_to_not};
use hittingset::{DefaultEdge, Edge, Hypergraph, ImplicableEdge, NodeType};
use crate::problems::results::MultiFD;

type Edgemark = FixedBitSet;


pub struct DynamicEdgemarkStoreMulti<'a> {
    graph: &'a mut Hypergraph<MultiFD>,
    vertex_hittings: Vec<Edgemark>,
    critical: Vec<Vec<Edgemark>>,
    uncov: Edgemark,
    nodes: Vec<NodeType>,
    cand: Vec<MultiFD>,
    current_vertex_hitting: Edgemark,
}

impl<'a> DynamicEdgemarkStoreMulti<'a> {
    pub fn get_nodes(&self) -> &[NodeType] {
        &self.nodes
    }
}

impl<'a> DynamicEdgemarkStoreMulti<'a> {
    pub fn new(graph: &'a mut Hypergraph<MultiFD>) -> DynamicEdgemarkStoreMulti<'a> {
        let vertex_count = graph.vertex_count();
        let edge_count = graph.edge_count();
        let mut critical = Vec::with_capacity(vertex_count);
        for i in 0..vertex_count {
            critical.push(vec![Edgemark::with_capacity(edge_count); i]);
        }

        let vertex_hittings = vec!(Edgemark::with_capacity(edge_count); vertex_count / 2);

        let cand = vec!(MultiFD::filled(vertex_count); vertex_count);

        let mut uncov = Edgemark::with_capacity(edge_count);
        uncov.set_range(0..edge_count, true);

        let mut res = DynamicEdgemarkStoreMulti {
            cand,
            graph,
            critical,
            uncov,
            vertex_hittings,
            nodes: Vec::with_capacity(vertex_count),
            current_vertex_hitting: Edgemark::with_capacity(edge_count),
        };
        res.update_vertex_hittings(0);
        res
    }

    pub fn is_fully_covered(&self) -> bool {
        self.uncov.ones().next() == None
    }

    pub fn uncovered_edges(&self) -> impl Iterator<Item=&MultiFD> {
        self.uncov.ones().map(|idx| self.graph.get_edge(idx))
    }

    pub fn get_cand(&self) -> &[MultiFD] {
        &self.cand
    }

    pub fn get_current_cand(&self) -> &MultiFD {
        &self.cand[self.nodes.len()]
    }

    pub fn get_current_cand_mut(&mut self) -> &mut MultiFD {
        &mut self.cand[self.nodes.len()]
    }

    fn current_criticals(&self) -> &[Edgemark] {
        &self.critical[self.nodes.len()]
    }

    fn can_add(&self, vertex: usize) -> bool {
        !self.current_criticals().iter().enumerate().any(|(idx, em) | {
            match self.nodes[idx] {
                NodeType::LHS(_) => em.is_subset(&self.vertex_hittings[vertex]),
                _ => false
            }
        })
    }

    fn can_add_rhs(&self, vertex: usize) -> bool {
        !self.current_criticals().iter().enumerate().any(|(idx, em) | {
            match self.nodes[idx] {
                NodeType::LHS(_) => is_subset_inverse(em, &self.vertex_hittings[vertex]),
                _ => false
            }
        })
    }

    pub fn add_lhs(&mut self, lhs: usize) -> bool {
       if !self.can_add(lhs) {
           return false;
       }

        let cand = self.copy_last_cand();
        cand.remove(lhs);
        cand.rhs.remove(lhs);
        if cand.rhs.fast_nodes().next().is_none() {
            return false;
        }
        self.nodes.push(NodeType::LHS(lhs));

        Self::add_crit_level(&mut self.critical, &self.vertex_hittings[lhs], &mut self.uncov, self.nodes.len());

        self.check_criticality_for_active_rhs();

        self.remove_lhs_for_singular_rhs();

        true
    }

    fn copy_last_cand(&mut self) -> &mut MultiFD {
        let (old, new) = self.cand.split_at_mut(self.nodes.len() + 1);
        new[0].set_to(&old[self.nodes.len()]);
        &mut new[0]
    }

    pub fn add_rhs(&mut self, rhs: &DefaultEdge) -> bool {
        let cand = self.copy_last_cand();
        cand.rhs.intersect_with(rhs);

        self.nodes.push(NodeType::RHS);
        self.calc_rhs_hitting();

        Self::add_crit_level(&mut self.critical, &self.current_vertex_hitting, &mut self.uncov, self.nodes.len());

        self.remove_lhs_for_singular_rhs();


        true
    }

    fn calc_rhs_hitting(&mut self) {
        let mut iter = self.cand[self.nodes.len()].rhs.fast_nodes();
        set_to_not(&mut self.current_vertex_hitting, &self.vertex_hittings[iter.next().unwrap()]);
        for column in iter {
            intersect_with_inverse(&mut self.current_vertex_hitting, &self.vertex_hittings[column]);
        }
    }

    fn check_criticality_for_active_rhs(&mut self) {
        for rhs in self.get_current_cand().rhs.clone().fast_nodes() {
            if !self.can_add_rhs(rhs) {
                self.get_current_cand_mut().rhs.remove(rhs);
            }
        }
    }

    fn remove_lhs_for_singular_rhs(&mut self) {
        let mut iter = self.get_current_cand().rhs.fast_nodes();
        if let (Some(rhs), None) = (iter.next(), iter.next()) {
            self.get_current_cand_mut().lhs.remove(rhs);
        }
    }

    fn add_crit_level(critical: &mut Vec<Vec<Edgemark>>, current: &Edgemark, uncov: &mut Edgemark, pos: usize) {
        // if an edge is in the current vertex hitting, it can no longer be critical
        let split = critical.split_at_mut(pos);
        for (old, new) in split.0[pos - 1].iter().zip(split.1[0].iter_mut()) {
            set_to(new, old);
            new.difference_with(current);
        }
        // all previously uncovered edges that are now covered are critical for v
        let last = critical[pos].last_mut().unwrap();
        set_to(last, current);
        *last &= &*uncov;

        uncov.difference_with(current);
    }

    pub fn pop(&mut self) {
        let lhs = self.nodes.pop().unwrap();
        if let NodeType::LHS(node) = lhs {
            self.get_current_cand_mut().add(node);
        }

        self.uncov.union_with(self.critical[self.nodes.len() + 1].last().unwrap());
    }

    pub fn extend(&mut self, s: &MultiFD, new_edges: Hypergraph<MultiFD>) {
        let previous_size = self.graph.edge_count();

        self.graph.reserve(new_edges.edge_count());
        for e in new_edges {
            if !self.graph.implies(&e) {
                self.graph.add_edge(e);
            }
        }

        if previous_size == self.graph.edge_count() { return; }
        for v in self.vertex_hittings.iter_mut() {
            v.grow(self.graph.edge_count());
        }
        self.update_vertex_hittings(previous_size);

        self.uncov.grow(self.graph.edge_count());

        for crits in self.critical.iter_mut() {
            for crit in crits.iter_mut() {
                crit.grow(self.graph.edge_count());
            }
        }

        for (idx, edge) in self.graph.iter().enumerate().skip(previous_size) {
            if !s.covers(edge) {
                self.uncov.set(idx, true);
            } else {
                //println!("Not disjoint!");

                let mut iter = (0..self.nodes.len()).filter(|i| {
                    match &self.nodes[*i] {
                        NodeType::LHS(lhs) => edge.contains(*lhs) || self.cand[*i + 1].rhs.implies(&edge.rhs),
                        NodeType::RHS => self.cand[*i + 1].rhs.implies(&edge.rhs)
                    }
                });
                let first = iter.next().unwrap();
                let end = iter.next().unwrap_or(self.nodes.len());
                for i in first..end {
                    self.critical[i + 1][first].set(idx, true);
                }
            }
        }
    }

    fn update_vertex_hittings(&mut self, offset: usize) {
        for (idx, edge) in self.graph.iter().enumerate().skip(offset) {
            for v in edge.lhs.fast_nodes() {
                self.vertex_hittings[v].set(idx, true);
            }
        }
    }
}
