use std::borrow::BorrowMut;
use std::cmp::min;
use serde::{Serialize, Deserialize};
use tracing::trace_span;
use hittingset::{Edge, NodeType};
use plistore::{Cluster, ClusterFilter, Pli, PLITable};
use std::cell::RefCell;
use std::fs::File;
use std::io::Write;
use std::marker::PhantomData;
use crate::sampling::SamplingStrategy;

pub struct PLICache<'a, E : Edge + Clone + Serialize, S: SamplingStrategy<E>> {
    intersections: Vec<Pli>,
    cached: usize,
    table: &'a PLITable,
    sampling: &'a S,
    phantom: PhantomData<E>
}

#[derive(Debug)]
#[derive(Serialize, Deserialize)]
pub enum PLICacheOps<E> {
    GET_PLI(Vec<NodeType>, Vec<E>),
    TRUNCATE(usize),
}

#[cfg(feature = "collect_pli_profiles")]
thread_local! {
    pub static PROFILE_FILE : RefCell<Option<File>> = RefCell::new(None);
}

#[cfg(feature = "collect_pli_profiles")]
fn write<E : Serialize>(op : &PLICacheOps<E>) {
    PROFILE_FILE.with(|mut f| f.borrow_mut().as_mut().unwrap().write_fmt(format_args!("{}\n", serde_json::to_string(op).unwrap()))).expect("Failed to write profile");
}


impl<'a, E : Edge + Clone + Serialize, S: SamplingStrategy<E>> PLICache<'a, E, S> {
    pub fn new(table: &'a PLITable, sampling: &'a S) -> PLICache<'a, E, S> {
        PLICache {
            intersections: vec![Pli::new(); table.column_count() * 2],
            cached: 0,
            table,
            sampling,
            phantom: Default::default()
        }
    }

    pub fn get_pli(&mut self, columns: &[NodeType], filter: &mut impl ClusterFilter, cand: &[E]) -> &[Cluster] {
        #[cfg(feature = "collect_pli_profiles")]
        {
            write(&PLICacheOps::GET_PLI(columns.to_owned(), Vec::from(&cand[..columns.len() + 1])));
        }

        let column_count = columns.len();
        if column_count == 0 {
            self.table.get_full_pli()
        } else if column_count == 1 {
            match columns[0] {
                NodeType::LHS(lhs) => self.table.get_column_pli(lhs),
                NodeType::RHS => self.table.get_full_pli()
            }
        } else {
            self.pull_up_intersections(columns, cand, filter);
            &self.intersections[column_count - 2]
        }
    }

    pub fn truncate(&mut self, current: usize) {
        #[cfg(feature = "collect_pli_profiles")]
        {
            write::<E>(&PLICacheOps::TRUNCATE(current));
        }

        self.cached = min(self.cached, current);
    }

    fn pull_up_intersections<F: ClusterFilter>(&mut self, columns: &[NodeType], cand: &[E], filter: &mut F) {
        // let span = span!(Level::TRACE, "find_common");
        // let enter = span.enter();
        // for _ in 0..self.column_indices.len() {
        //     if let Some(NodeType::RHS(rhs)) = iter.next() {
        //         current_filter.set_rhs(rhs);
        //     }
        // }
        // drop(enter);

        for addition in &columns[self.cached..] {
            self.cached += 1;

            if self.cached <= 1 {
                continue;
            }

            //let span = span!(Level::TRACE, "prelude");
            //let enter = span.enter();
            let mut current_filter = SamplingStrategyFilter {
                candidate: &cand[self.cached],
                sampling: self.sampling,
                filter,
            };

            //drop(enter);
            //let mut current_filter = &mut *filter;

            let intersection_idx = self.cached - 2;

            let split = self.intersections.split_at_mut(intersection_idx);
            let current = if intersection_idx == 0 {
                match columns[0] {
                    NodeType::LHS(lhs) => self.table.get_column_pli(lhs),
                    NodeType::RHS => self.table.get_full_pli()
                }
            } else {
                &split.0[intersection_idx - 1]
            };

            split.1[0].clear();
            current_filter.set_requested_level(self.cached == columns.len());
            match addition {
                NodeType::LHS(lhs) => {
                    //let start = Instant::now();
                    self.table.intersect(current, intersection_idx == 0, *lhs, &mut current_filter, &mut split.1[0]);
                    // let time = Instant::now() - start;
                    // if time.as_micros() > 500 {
                    //     println!("{:?} {:?} {};  {} {} ->  {} {}", time, columns, addition, current.len(), get_size(current), split.1[0].len(), get_size(&split.1[0]));
                    // }

                },
                NodeType::RHS => {
                    filter_rhs(&mut current_filter, current, &mut split.1[0]);
                }
            }

            current_filter.set_requested_level(false);
        }
    }
}



pub struct SamplingStrategyFilter<'a, E:Edge, S:SamplingStrategy<E>, F : ClusterFilter> {
    candidate: &'a E,
    sampling: &'a S,
    filter: &'a mut F,
}


impl <'a, E:Edge, S:SamplingStrategy<E>, F : ClusterFilter> ClusterFilter for SamplingStrategyFilter<'a, E, S, F> {
    fn keep(&mut self, cluster: &Cluster) -> bool {
        self.sampling.get_violating_pair(cluster, self.candidate).is_some() && self.filter.keep(cluster)
    }

    fn set_requested_level(&mut self, active: bool) {
        self.filter.set_requested_level(active);
    }
}

fn filter_rhs<'a, E:Edge, S:SamplingStrategy<E>, F : ClusterFilter>(filter: &mut SamplingStrategyFilter<'a, E, S, F>, input: &[Cluster], output: &mut Pli) {
    trace_span!("filter_rhs").in_scope(|| {
        for cluster in input {
            if filter.keep(cluster) {
                output.push(cluster.clone());
            }
        }
    });

}