use std::hash::Hash;
use tracing::{Level, span};

use hittingset::{Edge, Hypergraph, ImplicableEdge, NodeType, Validator};
use plistore::{PLITable};

use crate::edgecollectors::{EdgeConsumer, EdgeHashCollector};
use crate::problems::ProblemConfig;
use crate::sampling::SamplingStrategy;
use crate::validation::clusterfilters::SmallClusterCollector;
use crate::validation::plicache::PLICache;

pub struct PLIValidator<'a, C: EdgeConsumer<P::EdgeType>, S: SamplingStrategy<P::EdgeType>, P: ProblemConfig> where P::EdgeType : Clone {
    consumer: &'a mut C,
    sampling: &'a S,
    problem: &'a P,
    cache: PLICache<'a, P::EdgeType, S>,
}

impl<'a, S: SamplingStrategy<P::EdgeType>, C: EdgeConsumer<P::EdgeType>, P: ProblemConfig> Validator<P::EdgeType> for PLIValidator<'a, C, S, P> where P::EdgeType : Hash + Eq + ImplicableEdge + Clone {
    fn check(&mut self, s: &P::EdgeType, columns: &[NodeType], cand: &[P::EdgeType]) -> (bool, Option<Hypergraph<P::EdgeType>>) {

        let mut set = EdgeHashCollector::new();
        let mut filter = SmallClusterCollector::new(self.sampling, &mut set, 100, true);

        let span = span!(Level::TRACE, "get_pli");
        let enter = span.enter();
        let pli = self.cache.get_pli(columns, &mut filter, cand);
        drop(enter);

        if let Some((p1,p2)) = pli.iter().find_map(|c|self.sampling.get_violating_pair(c, s)) {
            self.problem.get_edge(p1, p2, &mut set);
        }

        if set.get_edge_set().iter().all(|e| s.covers(e)) {
            if self.problem.is_valid_result(s) {
                self.consumer.consume(s.clone());
            }
            (false, Some(set.to_graph(s)))
        } else {
            let span = span!(Level::TRACE, "sample");
            let enter = span.enter();
            self.sampling.sample_pli(pli, &mut set);
            drop(enter);

            (true, Some(set.to_graph(s)))
        }
    }

    fn push(&mut self, columns: &[NodeType]) {
        if !columns.is_empty() {
            self.cache.truncate(columns.len() - 1);
        }
    }
}

impl<'a, S: SamplingStrategy<P::EdgeType>, C: EdgeConsumer<P::EdgeType>, P: ProblemConfig> PLIValidator<'a, C, S, P> where P::EdgeType : Clone {
    pub(crate) fn new(table: &'a PLITable, consumer: &'a mut C, problem: &'a P, sampling: &'a S) -> PLIValidator<'a, C, S, P> {
        PLIValidator {
            cache: PLICache::new(table, sampling),
            consumer,
            sampling,
            problem
        }
    }
}

