use std::iter::Iterator;
use std::time::Instant;

use arrow::array::{Array, ArrayRef};

use pyo3::prelude::*;
use libc::uintptr_t;
use pyo3::wrap_pyfunction;
use fddiscovery::edgecollectors::VecEdgeCollector;
use fddiscovery::problems::results::MultiFD;

use crate::errors::PyO3ArrowError;
use crate::preprocess::preprocess;

mod errors;
mod preprocess;

fn to_py(array: ArrayRef, py: Python) -> PyResult<PyObject> {
    let (array_pointer, schema_pointer) =
        array.to_raw().map_err(PyO3ArrowError::from)?;

    let pa = py.import("pyarrow")?;

    let array = pa.getattr("Array")?.call_method1(
        "_import_from_c",
        (array_pointer as uintptr_t, schema_pointer as uintptr_t),
    )?;
    Ok(array.to_object(py))
}

#[pyclass]
struct Fd {
    lhs: Vec<String>,
    rhs: String
}

#[pymethods]
impl Fd {

    #[getter]
    fn lhs(&self) -> PyResult<Vec<String>> {
        Ok(self.lhs.clone())
    }
    #[getter]
    fn rhs(&self) -> PyResult<&String> {
        Ok(&self.rhs)
    }

    fn __repr__(&self) -> String {
        format!("FD({} -> {})", self.lhs.join(","), self.rhs)
    }

    fn __str__(&self) -> String {
        format!("FD({} -> {})", self.lhs.join(","), self.rhs)
    }
}

#[pyfunction]
fn search(arrays: Vec<PyObject>, py: Python) -> PyResult<Vec<Fd>> {

    let preprocessing_start = Instant::now();
    let (columns, mut plis) = preprocess(arrays, py)?;

    let preprocessing_end = Instant::now();
    println!("Preprocessing: {:?}", preprocessing_end - preprocessing_start);

    let discovery_start = Instant::now();
    let mut result = Vec::new();
    let mut collector1 : VecEdgeCollector<MultiFD> =  VecEdgeCollector::new();
    fddiscovery::problems::fds::search_fds(&mut plis, false, &mut collector1);
    for fd in &collector1.results {
        result.push(Fd {
            lhs: fd.lhs.fast_nodes().map(|idx| columns[idx].to_string()).collect(),
            rhs: fd.rhs.fast_nodes().map(|idx| columns[idx].to_string()).next().unwrap(),
        })
    }
    let discovery_end = Instant::now();
    println!("Discovery: {:?}", discovery_end - discovery_start);

    // export
    Ok(result)
}


/// A Python module implemented in Rust.
#[pymodule]
fn fddiscovery(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(search, m)?)?;
    m.add_class::<Fd>()?;
    Ok(())
}