use std::collections::HashMap;
use std::hash::Hash;
use arrow::array::{Array, ArrayRef, make_array_from_raw, StructArray};
use arrow::datatypes::DataType;
use arrow::error::ArrowError;
use libc::uintptr_t;
use pyo3::prelude::*;
use plistore::{Cluster, PLITable, RowIndex};
use crate::PyO3ArrowError;

pub(crate) fn preprocess(arrays: Vec<PyObject>, py: Python) -> PyResult<(Vec<String>, PLITable)> {
    let rust_arrays2 : Vec<ArrayRef> = arrays.into_iter().map(|array| {
        to_rust(array, py)
    }).collect::<PyResult<Vec<_>>>()?;

    let rust_arrays = rust_arrays2.iter().map(|array| {
        array.as_any()
            .downcast_ref::<StructArray>()
            .ok_or_else(|| PyO3ArrowError::ArrowError(ArrowError::ParseError(
                "Expects a StructRef".to_string(),
            )))
            .map_err(PyErr::from)
    }).collect::<PyResult<Vec<_>>>()?;

    let schema = rust_arrays.iter().next().unwrap();

    let mut plis = PLITable::new(rust_arrays.iter().map(|a| a.len() as RowIndex).sum(), schema.num_columns());


    for i in 0..schema.num_columns() {
        let c = schema.column(i);
        match c.data_type() {
            DataType::Boolean => build_index::<_, arrow::array::BooleanArray>(&rust_arrays, i, &mut plis),
            DataType::Int8 => build_index::<_, arrow::array::Int8Array>(&rust_arrays, i, &mut plis),
            DataType::Int16 => build_index::<_, arrow::array::Int16Array>(&rust_arrays, i, &mut plis),
            DataType::Int32 => build_index::<_, arrow::array::Int32Array>(&rust_arrays, i, &mut plis),
            DataType::Int64 => build_index::<_, arrow::array::Int64Array>(&rust_arrays, i, &mut plis),
            DataType::UInt8 => build_index::<_, arrow::array::UInt8Array>(&rust_arrays, i, &mut plis),
            DataType::UInt16 => build_index::<_, arrow::array::UInt16Array>(&rust_arrays, i, &mut plis),
            DataType::UInt32 => build_index::<_, arrow::array::UInt32Array>(&rust_arrays, i, &mut plis),
            DataType::UInt64 => build_index::<_, arrow::array::UInt64Array>(&rust_arrays, i, &mut plis),
            DataType::Date32 => build_index::<_, arrow::array::Date32Array>(&rust_arrays, i, &mut plis),
            DataType::Date64 => build_index::<_, arrow::array::Date64Array>(&rust_arrays, i, &mut plis),
            DataType::Time32(_) => build_index::<_, arrow::array::Time32MillisecondArray>(&rust_arrays, i, &mut plis),
            DataType::Time64(_) => build_index::<_, arrow::array::Time64NanosecondArray>(&rust_arrays, i, &mut plis),
            DataType::Decimal(_, _) => build_index::<_, arrow::array::DecimalArray>(&rust_arrays, i, &mut plis),
            DataType::Utf8 => build_index::<_, arrow::array::StringArray>(&rust_arrays, i, &mut plis),
            DataType::LargeUtf8 => build_index::<_, arrow::array::LargeStringArray>(&rust_arrays, i, &mut plis),
            _ => println!("Unsupported! {}", c.data_type()),
        };
    }

    Ok((schema.column_names().into_iter().map(String::from).collect(), plis))
}


fn build_index<'a, T,V>(rust_arrays: &std::vec::Vec<&'a StructArray>, i: usize, plis: &mut PLITable) where T: Hash + Eq, V: 'static, &'a  V: IntoIterator<Item = Option<T>> + 'a  {
    let mut map: HashMap<Option<T>, Cluster> = HashMap::new();
    let mut rownumber : RowIndex = 0;
    rust_arrays.iter().for_each(|array| {
        for x in array.column(i).as_any().downcast_ref::<V>().unwrap().into_iter() {
            map.entry(x).or_default().push(rownumber);
            rownumber += 1;
        }
    });
    map.into_values().filter(|c| c.len() > 1).for_each(|cluster| {
        plis.add_cluster(i, cluster);
    });
}


fn to_rust(ob: PyObject, py: Python) -> PyResult<ArrayRef> {
    // prepare a pointer to receive the Array struct
    let (array_pointer, schema_pointer) =
        arrow::ffi::ArrowArray::into_raw(unsafe { arrow::ffi::ArrowArray::empty() });

    // make the conversion through PyArrow's private API
    // this changes the pointer's memory and is thus unsafe. In particular, `_export_to_c` can go out of bounds
    ob.call_method1(
        py,
        "_export_to_c",
        (array_pointer as uintptr_t, schema_pointer as uintptr_t),
    )?;
    let array = unsafe { make_array_from_raw(array_pointer, schema_pointer) }
        .map_err(PyO3ArrowError::from)?;
    Ok(array)
}