use std::borrow::BorrowMut;
#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};
use std::cell::RefCell;
use std::fs::File;
use std::io::Write;
use crate::{Edge, Hypergraph, ImplicableEdge};
use crate::edgemark::Edgemark;
use crate::treesearch::NodeType;

#[cfg(feature = "collect_profiles")]
thread_local! {
    pub static PROFILE_FILE : RefCell<Option<File>> = RefCell::new(None);
}

pub struct DynamicEdgemarkStore<'a, E: Edge> {
    graph: &'a mut Hypergraph<E>,
    vertex_hittings: Vec<Edgemark>,
    critical: Vec<Vec<Edgemark>>,
    uncov: Edgemark,
    nodes: Vec<NodeType>,
    cand: Vec<E>,
}

impl<'a, E: Edge> DynamicEdgemarkStore<'a, E> {
    pub(crate) fn get_nodes(&self) -> &[NodeType] {
        &self.nodes
    }
}

#[derive(Debug)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
pub enum StoreOps<E: Edge> {
    New(Hypergraph<E>),
    Add(usize),
    Pop,
    IsFullyCovered,
    UncoveredEdges,
    Extend(Hypergraph<E>),
}


#[cfg(feature = "collect_profiles")]
fn write<E : Edge>(op : &StoreOps<E>) {
    PROFILE_FILE.with(|f| f.borrow_mut().as_mut().unwrap().write_fmt(format_args!("{}\n", serde_json::to_string(op).unwrap()))).expect("Failed to write profile");
}


impl<'a, E: Edge + ImplicableEdge + Clone> DynamicEdgemarkStore<'a, E> {
    pub fn new(graph: &'a mut Hypergraph<E>) -> DynamicEdgemarkStore<'a, E> {
        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::New(graph.clone()));
        }

        let vertex_count = graph.vertex_count();
        let edge_count = graph.edge_count();
        let mut critical = Vec::with_capacity(vertex_count);
        for i in 0..vertex_count {
            critical.push(vec![Edgemark::empty(edge_count); i]);
        }

        let vertex_hittings = vec!(Edgemark::empty(edge_count); vertex_count);

        let cand = vec!(E::filled(vertex_count); vertex_count);

        let mut uncov = Edgemark::filled(edge_count);

        let mut res = DynamicEdgemarkStore {
            cand,
            graph,
            critical,
            uncov,
            vertex_hittings,
            nodes: Vec::with_capacity(vertex_count),
        };
        res.update_vertex_hittings(0);
        res
    }

    pub fn is_fully_covered(&self) -> bool {
        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::<E>::IsFullyCovered);
        }

        self.uncov.is_empty()
    }

    pub fn uncovered_edges(&self) -> impl Iterator<Item=&E> {
        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::<E>::UncoveredEdges);
        }
        self.uncov.ones().map(|idx| self.graph.get_edge(idx))
    }

    pub(crate) fn get_cand(&self) -> &[E] { &self.cand }

    pub fn get_current_cand(&self) -> &E {
        &self.cand[self.nodes.len()]
    }

    pub(crate) fn get_current_cand_mut(&mut self) -> &mut E {
        &mut self.cand[self.nodes.len()]
    }

    fn current_criticals(&self) -> &[Edgemark] {
        &self.critical[self.nodes.len()]
    }

    pub fn add(&mut self, vertex: usize) -> bool {
        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::<E>::Add(vertex));
        }
        if self.current_criticals().iter().any(|em| em.is_subset(&self.vertex_hittings[vertex])) {
            return false;
        }

        self.nodes.push(NodeType::LHS(vertex));

        let (old, new) = self.cand.split_at_mut(self.nodes.len());
        new[0].set_to(&old[self.nodes.len() - 1]);
        new[0].remove(vertex);

        self.add_crit_level(vertex);

        self.uncov -= &self.vertex_hittings[vertex];
        true
    }

    fn add_crit_level(&mut self, vertex: usize) {
        let vertex_hitting = &self.vertex_hittings[vertex];

        // if an edge is in the current vertex hitting, it can no longer be critical
        let split = self.critical.split_at_mut(self.nodes.len());
        for (old, new) in split.0[(self.nodes.len() - 1)].iter().zip(split.1[0].iter_mut()) {
            new.clone_from(old);
            *new -= vertex_hitting;
        }
        // all previously uncovered edges that are no covered are critical for v
        let last = self.critical[self.nodes.len()].last_mut().unwrap();
        last.clone_from(vertex_hitting);
        *last &= &self.uncov;
    }

    pub fn pop(&mut self) {
        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::<E>::Pop);
        }
        if let NodeType::LHS(removed) = self.nodes.pop().unwrap() {
            self.get_current_cand_mut().add(removed);
        }

        self.uncov |= self.critical[self.nodes.len() + 1].last().unwrap();
    }

    pub fn extend(&mut self, s: &E, new_edges: Hypergraph<E>) {
        if new_edges.edge_count() == 0 {
            return;
        }

        #[cfg(feature = "collect_profiles")]
        {
            write(&StoreOps::<E>::Extend(new_edges.clone()));
        }
        let previous_size = self.graph.edge_count();

        self.graph.reserve(new_edges.edge_count());
        for e in new_edges {
            if !self.graph.implies(&e) {
                self.graph.add_edge(e);
            }
        }

        if previous_size == self.graph.edge_count() { return; }
        for v in self.vertex_hittings.iter_mut() {
            v.grow(self.graph.edge_count());
        }
        self.update_vertex_hittings(previous_size);

        self.uncov.grow(self.graph.edge_count());

        for crits in self.critical.iter_mut() {
            for crit in crits.iter_mut() {
                crit.grow(self.graph.edge_count());
            }
        }

        for (idx, edge) in self.graph.iter().enumerate().skip(previous_size) {
            if !s.covers(edge) {
                self.uncov.set(idx);
            } else {
                let mut iter = (0..self.nodes.len()).filter(|i| {
                    match self.nodes[*i] {
                        NodeType::LHS(lhs) => edge.contains(lhs),
                        _ => false
                    }
                });
                let first = iter.next().unwrap();
                let end = iter.next().unwrap_or(self.nodes.len());
                for i in first..end {
                    self.critical[i + 1][first].set(idx);
                }
            }
        }
    }

    fn update_vertex_hittings(&mut self, offset: usize) {
        for (idx, edge) in self.graph.iter().enumerate().skip(offset) {
            for v in edge.nodes() {
                self.vertex_hittings[v].set(idx);
            }
        }
    }
}
