use std::fmt;
use std::fmt::Formatter;
use fixedbitset::{FixedBitSet, Ones};
use itertools::Itertools;
#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};



use crate::bitutil;
use crate::bitutil::get_intersection_size;

#[cfg(feature = "serde")]
pub trait Edge: Ord + Serialize  {
    fn filled(size: usize) -> Self;
    fn empty(size: usize) -> Self;
    fn nodes(&'_ self) -> Box<dyn Iterator<Item=usize> + '_>;
    fn add(&mut self, node: usize);
    fn remove(&mut self, node: usize);
    fn overlap(&self, other: &Self) -> usize;
    fn covers(&self, other: &Self) -> bool;
    fn remove_all(&mut self, other: &Self);
    fn add_all(&mut self, other: &Self);
    fn intersect(&self, other: &Self) -> Self;
    fn remove_all_invalid(&mut self, other: &Self);
    fn set_to(&mut self, other: &Self);

    fn contains(&self, node: usize) -> bool;
    fn vertex_count(&self) -> usize;
    fn len(&self) -> usize;
}

#[cfg(not(feature = "serde"))]
pub trait Edge: Ord  {
    fn filled(size: usize) -> Self;
    fn empty(size: usize) -> Self;
    fn nodes(&'_ self) -> Box<dyn Iterator<Item=usize> + '_>;
    fn add(&mut self, node: usize);
    fn remove(&mut self, node: usize);
    fn overlap(&self, other: &Self) -> usize;
    fn is_disjoint(&self, other: &Self) -> bool;
    fn remove_all(&mut self, other: &Self);
    fn add_all(&mut self, other: &Self);
    fn intersect(&self, other: &Self) -> Self;
    fn remove_all_invalid(&mut self, other: &Self);
    fn set_to(&mut self, other: &Self);

    fn contains(&self, node: usize) -> bool;
    fn vertex_count(&self) -> usize;
    fn len(&self) -> usize;
}

pub trait ImplicableEdge : Sized {
    fn implies(&self, other: &Self) -> bool;
    fn is_implied_by(&self, others: &[Self]) -> bool;
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Ord, PartialOrd)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
pub struct DefaultEdge(FixedBitSet);

impl DefaultEdge {
    pub fn clear(&mut self) {
        self.0.clear();
    }
}

impl DefaultEdge {
    pub fn intersect_with(&mut self, other: &Self) {
        self.0.intersect_with(&other.0);
    }

    pub fn fast_nodes(&self) -> Ones {
        self.0.ones()
    }
}

impl Edge for DefaultEdge {
    fn filled(size: usize) -> Self {
        let mut res = FixedBitSet::with_capacity(size);
        res.set_range(0..size, true);
        DefaultEdge(res)
    }

    fn empty(size: usize) -> Self {
        DefaultEdge(FixedBitSet::with_capacity(size))
    }

    fn nodes(&self) -> Box<dyn Iterator<Item=usize> + '_> {
        Box::new(self.0.ones())
    }

    fn add(&mut self, node: usize) {
        self.0.set(node, true);
    }

    fn remove(&mut self, node: usize) {
        self.0.set(node, false);
    }


    fn overlap(&self, other: &Self) -> usize {
        get_intersection_size(&self.0, &other.0)
    }

    fn covers(&self, other: &Self) -> bool {
        !self.0.is_disjoint(&other.0)
    }

    fn remove_all(&mut self, other: &Self) {
        self.0.difference_with(&other.0);
    }


    fn add_all(&mut self, other: &Self) {
        self.0 |= &other.0;
    }

    fn intersect(&self, other: &Self) -> Self {
        DefaultEdge(&self.0 & &other.0)
    }

    fn remove_all_invalid(&mut self, _: &Self) {}

    fn contains(&self, node: usize) -> bool { self.0.contains(node) }
    fn vertex_count(&self) -> usize {
        self.0.count_ones(..)
    }
    fn len(&self) -> usize {
        self.0.len()
    }

    fn set_to(&mut self, other: &Self) {
        bitutil::set_to(&mut self.0, &other.0);
    }
}


impl ImplicableEdge for DefaultEdge {
    fn implies(&self, other: &Self) -> bool {
        self.0.is_subset(&other.0)
    }

    fn is_implied_by(&self, others: &[Self]) -> bool {
        others.iter().any(|e| e.implies(self))
    }
}

impl fmt::Display for DefaultEdge {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.fast_nodes().join(","))
    }
}