use std::slice::Iter;
use std::vec::IntoIter;
#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

use crate::edge::{Edge, ImplicableEdge};

#[derive(Debug, Clone)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
pub struct Hypergraph<E: Edge> {
    vertices_count: usize,
    edges: Vec<E>,
}

impl<E: Edge> Hypergraph<E> {
    pub fn new(vertex_count: usize) -> Hypergraph<E> {
        Hypergraph {
            vertices_count: vertex_count,
            edges: Vec::new(),
        }
    }

    pub fn add_edge(&mut self, edge: E) { self.edges.push(edge); }
    pub fn get_edge(&self, idx: usize) -> &E { &self.edges[idx] }
    pub fn reserve(&mut self, size: usize) {
        self.edges.reserve(size);
    }

    pub fn vertex_count(&self) -> usize { self.vertices_count }
    pub fn edge_count(&self) -> usize { self.edges.len() }

    pub fn iter(&self) -> Iter<'_, E> { self.edges.iter() }
}

impl<E: Edge> IntoIterator for Hypergraph<E> {
    type Item = E;
    type IntoIter = IntoIter<E>;

    fn into_iter(self) -> Self::IntoIter {
        self.edges.into_iter()
    }
}

impl<'a, E: Edge> IntoIterator for &'a Hypergraph<E> {
    type Item = &'a E;
    type IntoIter = Iter<'a, E>;

    fn into_iter(self) -> Self::IntoIter {
        self.edges.iter()
    }
}

impl<E: Edge + ImplicableEdge> Hypergraph<E> {
    #[inline]
    pub fn add_edge_and_minimize(&mut self, edge: E) {
        if !self.implies(&edge) {
            self.edges.retain(|e| !edge.implies(&e));
            self.add_edge(edge);
        }
    }

    pub fn implies(&self, edge: &E) -> bool {
        edge.is_implied_by(&self.edges)
    }
}
