use std::fmt;
use std::fmt::Formatter;
#[cfg(test)]
use mockall::*;
#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

use crate::{Edge, Hypergraph, ImplicableEdge};
use crate::dynamicedgemarkstore::DynamicEdgemarkStore;

#[derive(Debug, PartialEq, Eq, Clone)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
pub enum NodeType {
    LHS(usize),
    RHS
}

impl fmt::Display for NodeType {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            NodeType::LHS(lhs) => write!(f, "LHS={}", lhs),
            NodeType::RHS => write!(f, "RHS")
        }

    }
}

#[cfg_attr(test, automock)]
pub trait Validator<E: Edge> {
    fn check(&mut self, candidate: &E, columns: &[NodeType], cand: &[E]) -> (bool, Option<Hypergraph<E>>);
    fn push(&mut self, columns: &[NodeType]);
}

pub fn search<E: Edge + Clone + ImplicableEdge>(partial: &mut Hypergraph<E>, target: &mut impl Validator<E>) {
    let vertex_count = partial.vertex_count();
    let mut s = E::empty(vertex_count);

    let mut store = DynamicEdgemarkStore::new(partial);

    extend_or_confirm(&mut s, &mut store, target);
}

fn extend_or_confirm<'a, E: Edge + Clone + ImplicableEdge>(s: &mut E, store: &mut DynamicEdgemarkStore<'a, E>, target: &mut impl Validator<E>) {
    target.push(store.get_nodes());

    if store.is_fully_covered() {
        let (cont, edges) = target.check(s, store.get_nodes(), store.get_cand());

        if let Some(new_edges) = edges {
            store.extend(s, new_edges);
        }
        if !cont {
            return;
        }
    }

    // find edge from uncov with smallest intersection c with cand
    let c = get_best_edge(store);

    store.get_current_cand_mut().remove_all(&c);

    for v in c.nodes() {
        if !store.add(v) {
            continue;
        }

        s.add(v);
        store.get_current_cand_mut().remove_all_invalid(s);

        extend_or_confirm(s, store, target);

        s.remove(v);

        store.pop();
    }
    store.get_current_cand_mut().add_all(&c);
}

#[inline]
fn get_best_edge<'a, E: Clone + Edge + ImplicableEdge>(edgemarkstore: &'a DynamicEdgemarkStore<'a, E>) -> E {
    edgemarkstore.uncovered_edges().min_by_key(|c| c.overlap(edgemarkstore.get_current_cand())).unwrap().intersect(edgemarkstore.get_current_cand())
}

#[cfg(test)]
mod tests {
    use crate::DefaultEdge;

    // Note this useful idiom: importing names from outer (for mod tests) scope.
    use super::*;

    macro_rules! edge {
        ( $size:expr; $( $x:expr ),* ) => {
            {
                let mut edge1 = DefaultEdge::empty($size);
                $(
                    edge1.add($x);
                )*
                edge1
            }
        };
    }

    macro_rules! result {
        ( $mock: expr; $( $x:expr ),* ; $result:expr ) => {
            {
                $mock.expect_check()
                .withf(|edge: &DefaultEdge, _, _| {
                    let target = [$( $x as usize ),*];
                    edge.nodes().eq(target.iter().copied())
                })
                .times(1)
                .return_const($result);
            }
        };
    }

    #[test]
    fn test_simple() {
        let mut g = Hypergraph::new(2);
        g.add_edge(edge!(2; 0));

        let mut mock = MockValidator::new();
        result!(mock; 0; (false, None));
        mock.expect_push().return_const(());

        search(&mut g, &mut mock);
    }

    #[test]
    fn test_simple_2() {
        let mut g = Hypergraph::new(5);
        g.add_edge(edge!(5; 0));
        g.add_edge(edge!(5; 1));

        let mut mock = MockValidator::new();
        result!(mock; 0, 1; (false, None));
        mock.expect_push().return_const(());

        search(&mut g, &mut mock);
    }

    #[test]
    fn test_minimality_check() {
        let mut g = Hypergraph::new(4);
        g.add_edge(edge!(4; 0, 2, 3));
        g.add_edge(edge!(4; 0, 1));

        let mut mock = MockValidator::new();
        result!(mock; 0; (false, None));
        result!(mock; 1, 2; (false, None));
        result!(mock; 1, 3; (false, None));
        mock.expect_push().return_const(());

        search(&mut g, &mut mock);
    }

    #[test]
    fn test_minimality_incomplete() {
        let mut g = Hypergraph::new(4);
        g.add_edge(edge!(4; 0, 2, 3));
        g.add_edge(edge!(4; 0, 1));


        let mut g2 = Hypergraph::new(4);
        g2.add_edge(edge!(4; 1, 2, 3));

        let mut mock = MockValidator::new();
        result!(mock; 0; (true, Some(g2)));
        result!(mock; 0, 1; (false, None));
        result!(mock; 1, 2; (false, None));
        result!(mock; 1, 3; (false, None));
        result!(mock; 0, 2; (false, None));
        result!(mock; 0, 3; (false, None));
        mock.expect_push().return_const(());

        search(&mut g, &mut mock);
    }
}