use std::cell::RefCell;
use tracing::{trace_span};

use crate::{Cluster, ClusterFilter, ClusterIndex, Pli, RefinementHelper, RowIndex};

thread_local! {
    static REFINEMENT_HELPER : RefCell<RefinementHelper> = RefCell::new(RefinementHelper::new());
}



pub struct PLITable {
    plis: Vec<Pli>,
    inverse: Vec<Vec<ClusterIndex>>,
    row_count: RowIndex,
    full_pli: Pli,
    name: Option<String>,
}

impl PLITable {
    pub fn new(row_count: RowIndex, column_count: usize, name: Option<String>) -> PLITable {
        PLITable {
            plis: vec![Pli::new(); column_count],
            inverse: vec![vec![0; row_count as usize]; column_count],
            row_count,
            full_pli: vec!((0..row_count as RowIndex).collect()),
            name
        }
    }

    pub fn get_name(&self) -> &Option<String> {
        &self.name
    }

    pub fn add_cluster(&mut self, column: usize, cluster: Cluster) {
        let counter = (self.plis[column].len() + 1) as ClusterIndex;
        for t in &cluster {
            self.inverse[column][*t as usize] = counter;
        }
        self.plis[column].push(cluster);
    }

    pub fn column_count(&self) -> usize {
        self.plis.len()
    }
    pub fn row_count(&self) -> RowIndex {
        self.row_count
    }

    pub fn get_column_pli(&self, column: usize) -> &[Cluster] {
        &self.plis[column]
    }
    pub fn get_full_pli(&self) -> &[Cluster] {
        &self.full_pli
    }

    pub fn is_different_on_column(&self, r1: RowIndex, r2: RowIndex, column: usize) -> bool {
        self.inverse[column][r1 as usize] == 0 || self.inverse[column][r1 as usize] != self.inverse[column][r2 as usize]
    }

    pub fn sort_plis_by(&mut self, target: usize) {
        sort_pli_by(&mut self.full_pli, &self.inverse[target]);
        self.plis.iter_mut().for_each(|pli| sort_pli_by(pli, &self.inverse[target]));
    }

    pub fn sort_pli_by(&self, pli: &mut Pli, target: usize) {
        sort_pli_by(pli, &self.inverse[target]);
    }

    pub fn intersect(&self, pli: &[Cluster], first: bool, column: usize, filter: &mut impl ClusterFilter, intersection: &mut Pli) {
        trace_span!("intersect").in_scope(|| {
            REFINEMENT_HELPER.with(|cell| {
                let mut refinement_helper = cell.borrow_mut();
                refinement_helper.reserve(self.row_count() as usize);
                for cluster in pli {
                    if first && !filter.keep(cluster) {
                        continue;
                    }

                    for &row in cluster {
                        if self.inverse[column][row as usize] != 0 {
                            refinement_helper.add(self.inverse[column][row as usize], row);
                        }
                    }
                    refinement_helper.add_refined_cluster(intersection, filter);
                }
            });
        })

    }
}

fn sort_pli_by(pli: &mut Pli, target: &[ClusterIndex]) {
    REFINEMENT_HELPER.with(|cell| {
        let mut refinement_helper = cell.borrow_mut();
        refinement_helper.reserve(target.len());
        for cluster in pli.iter_mut() {
            let first = target[*cluster.first().unwrap() as usize];
            if first == 0 || cluster.iter().any(|e| target[*e as usize] != first) {
                for row in cluster.iter() {
                    refinement_helper.add(target[*row as usize], *row);
                }
                refinement_helper.update_cluster(cluster);
            }
        }
    });
}

