use std::collections::HashMap;
use std::error::Error;
use std::fs::File;
use std::path::{Path, PathBuf};

use ahash::RandomState;
use csv::{ByteRecord, QuoteStyle, Reader, Writer};
use smallvec::SmallVec;
use serde::{Serialize, Deserialize};

use crate::{Cluster, RowIndex};
use crate::plis::PLITable;

#[derive(Default, Builder, Debug, Deserialize, Serialize, Clone)]
#[builder(setter(into))]
pub struct CSVConfig {
    #[builder(default = "true")]
    #[serde(default = "get_true")]
    header: bool,
    #[builder(default = "true")]
    #[serde(default = "get_true")]
    quoting: bool,
    #[builder(default = "Some(b'\\\\')")]
    #[serde(default = "default_escape")]
    escape: Option<u8>,
    #[builder(default = "true")]
    #[serde(default = "get_true")]
    double_quote: bool,
    #[builder(default = "b','")]
    #[serde(default = "default_delimiter")]
    delimiter: u8,
}

fn get_true() -> bool {
    true
}

fn default_delimiter() -> u8 {
    b','
}

fn default_escape() -> Option<u8> {
    Some(b'\\')
}

impl CSVConfig {
    fn to_essential(&self, path: &PathBuf) -> Result<EssentialTable, Box<dyn Error>> {
        let mut rdr = self.get_reader(path)?;
        let mut table = EssentialTable::new();
        table.set_name(path.file_name().unwrap().to_str().unwrap());
        let mut record = ByteRecord::new();
        let mut rownr = 0;
        while rdr.read_byte_record(&mut record)? {
            table.add_row(rownr, &record);
            rownr += 1;
        }
        Ok(table)
    }

    pub fn get_reader(&self, path: &Path) -> Result<Reader<File>, csv::Error> {
        csv::ReaderBuilder::new()
            .has_headers(self.header)
            .quoting(self.quoting)
            .escape(self.escape)
            .delimiter(self.delimiter)
            .double_quote(self.double_quote)
            .from_path(path)
    }

    pub fn get_writer(&self, path: &Path) -> Result<Writer<File>, csv::Error> {
        let mut rdr = csv::WriterBuilder::new();
        let mut rd = rdr.has_headers(self.header);
        if let Some(escape) = self.escape {
            rd = rd.escape(escape);
        }
        if self.quoting {
            rd = rd.quote_style(QuoteStyle::Necessary);
        } else {
            rd = rd.quote_style(QuoteStyle::Never);
        }

        rd = rd
            .delimiter(self.delimiter)
            .double_quote(self.double_quote);
        rd.from_path(path)
    }
}

struct EssentialTable {
    name: Option<String>,
    values: Vec<HashMap<SmallVec<[u8; 16]>, Cluster, RandomState>>,
    row_count: RowIndex,
}

impl EssentialTable {
    fn new() -> EssentialTable {
        EssentialTable { values: vec![], row_count: 0, name: None }
    }

    fn set_name(&mut self, name: &str) {
        self.name = Some(name.to_string());
    }

    fn add_row(&mut self, rownumber: RowIndex, record: &ByteRecord) {
        (1..=(record.len() - self.values.len())).for_each(|_| { self.values.push(HashMap::with_hasher(RandomState::new())) });
        for (column, value) in self.values.iter_mut().zip(record.iter()) {
            column.entry(SmallVec::from_slice(value)).or_default().push(rownumber);
        }
        self.row_count += 1;
    }

    fn into_table(self) -> PLITable {
        let mut result = PLITable::new(self.row_count, self.values.len(), self.name);
        for (column, clusters) in self.values.into_iter().enumerate() {
            clusters.into_values().filter(|c| c.len() > 1).for_each(|cluster| {
                result.add_cluster(column, cluster);
            });
        }
        result
    }
}

pub fn read_input(config: &CSVConfig, path: &PathBuf) -> Result<PLITable, Box<dyn Error>> {
    Ok(config.to_essential(path)?.into_table())
}
