use std::error::Error;
use std::fs::File;
use std::path::{Path, PathBuf};
use std::{fs, process};
use std::collections::HashSet;
use std::io::{LineWriter, Write};
use std::process::{Command};

use chrono::Utc;

use clap::{Parser};
use regex::Regex;
use serde::{Serialize, Deserialize};
use serde_json::Value;

use plistore::{CSVConfig};
use runner::sample::{ColumnSampler, DatasetSampler, RandomColumnSampler, RowSampler};

#[derive(Parser)]
#[clap(author, version, about, long_about = None)]
struct Cli {

    #[clap(short, long)]
    single: bool,

    /// Sets a custom config file
    #[clap(parse(from_os_str), value_name = "INPUT")]
    input: PathBuf,

    #[clap(short, long, default_value = "./hyfd.sh ./fdsingle.sh ./fdunion.sh")]
    commands: Vec<String>
}

#[derive(Default, Debug, Deserialize, Serialize)]
struct Experiment {
    #[serde(default = "default_timeout")]
    timeout: String,
    #[serde(default = "default_iterations")]
    iterations: usize,
    input: InputConfig,
    row_stride: Option<usize>,
    column_stride: Option<usize>,
}

fn default_timeout() -> String {
    String::from("1h")
}

fn default_iterations() -> usize {
    5
}


#[derive(Default, Debug, Deserialize, Serialize, Clone)]
pub struct InputConfig {
    files: Vec<PathBuf>,
    config: CSVConfig,
}

#[derive(Debug, Serialize, PartialEq)]
enum ExecutionStatus {
    Success,
    MemoryLimit,
    RuntimeLimit,
    Other(usize),
    Unknown
}

#[derive(Debug, Serialize)]
pub struct ExecutionResult {
    start: String,
    end: String,
    input: InputConfig,
    output: Option<Value>,
    algorithm: String,
    executionStatus: ExecutionStatus,
    memory: usize,
    stdout: String,
    stderr: String
}

fn run() -> Result<(), Box<dyn Error>> {

    let cli = Cli::parse();

    let output_dir = Path::new("output");


    fs::create_dir_all(output_dir);

    let output_path = output_dir.canonicalize().unwrap().join(format!("{}.json", Utc::now().to_rfc3339()));
    let mut writer = LineWriter::new(File::create(output_path)?);


    let temp_path = Path::new(cli.input.as_path().parent().unwrap()).join("temp/").join(cli.input.as_path().file_stem().unwrap());
    if !cli.single {
        fs::create_dir_all(&temp_path)?;
    }

    if cli.single {
        let input: InputConfig = serde_yaml::from_reader(File::open(&cli.input)?)?;
        let experiment = Experiment {
            timeout: default_timeout(),
            iterations: default_iterations(),
            input,
            row_stride: None,
            column_stride: None
        };
        for input_path in &experiment.input.files {
            let output_config = write_config(input_path.file_stem().unwrap().to_str().unwrap(), &temp_path, &experiment, format!("../../{}", input_path.to_str().unwrap()));
            run_iterations(&mut writer, &experiment, &output_config, &cli.commands);
        }
    } else {
        let experiment: Experiment = serde_yaml::from_reader(File::open(&cli.input)?)?;
        for input_path in &experiment.input.files {
            let input = cli.input.parent().unwrap().join(input_path);
            println!("{:?}", input);
            let row_count = experiment.input.config.get_reader(&input)?.into_byte_records().count();
            let column_count = experiment.input.config.get_reader(&input)?.into_byte_records().next().unwrap()?.len();


            if let Some(row_stride) = experiment.row_stride {
                for row_limit in (row_stride..=row_count).step_by(row_stride) {
                    let output_config = sample(format!("r{row_limit}"), &temp_path, &experiment, &input, RowSampler::new(row_limit));

                    run_iterations(&mut writer, &experiment, &output_config, &cli.commands);
                }
            }

            if let Some(column_stride) = experiment.column_stride {
                let mut current_algorithms : HashSet<&str> = HashSet::from_iter(cli.commands.iter().map(|s| s.as_ref()));
                for column_limit in (column_stride..=column_count).step_by(column_stride) {
                    let mut next_algorithms : HashSet<&str> = HashSet::new();
                    for _ in 0..experiment.iterations {
                        let output_config = sample(format!("c{column_limit}"), &temp_path, &experiment, &input, RandomColumnSampler::new(column_limit));
                        for command in &current_algorithms {
                            let result = run_experiment(&experiment.timeout, &output_config, command);

                            serde_json::to_writer(&mut writer, &result).expect("failed to write result");
                            writer.write_all(b"\n").expect("failed to write result");
                            writer.flush().expect("failed to write result");

                            if result.executionStatus == ExecutionStatus::Success {
                                next_algorithms.insert(*command);
                            }
                        }
                    }
                    current_algorithms = next_algorithms;

                }
            }
        }
    }


    Ok(())
}

fn sample(name: impl AsRef<str>, temp_path: &PathBuf, experiment: &Experiment, input: &PathBuf, sampler: impl DatasetSampler) -> PathBuf {
    let csv_name = format!("{}.csv", name.as_ref());
    let output = temp_path.join(&csv_name);
    sampler.sample(&input, &output, &experiment.input.config);

    // write input config
    write_config(name, temp_path, experiment, &csv_name)
}

fn write_config(name: impl AsRef<str>, temp_path: &PathBuf, experiment: &Experiment, csv_name: impl AsRef<str>) -> PathBuf {
    let new_config = InputConfig { files: vec![PathBuf::from(csv_name.as_ref())], config: experiment.input.config.clone() };
    let output_config = temp_path.join(format!("{}.yaml", name.as_ref()));
    serde_yaml::to_writer(File::create(&output_config).expect("failed to create config file"), &new_config).expect("failed to write config file");
    output_config
}

fn run_iterations(mut writer: &mut LineWriter<File>, experiment: &Experiment, output_config: &PathBuf, commands: &Vec<String>) {
    for command in commands {
        for _ in 0..experiment.iterations {
            let result = run_experiment(&experiment.timeout, &output_config, command);

            serde_json::to_writer(&mut writer, &result).expect("failed to write result");
            writer.write_all(b"\n").expect("failed to write result");
            writer.flush().expect("failed to write result");

            if result.executionStatus != ExecutionStatus::Success {
                break;
            }
        }
    }
}


fn run_experiment(timeout: &str, output_config: &PathBuf, command: &str) -> ExecutionResult {
    let start = Utc::now().to_rfc3339();

    let output_dir = Path::new("temp_output");

    fs::create_dir_all(output_dir);

    let output_path = output_dir.canonicalize().unwrap().join(format!("{}.json", Utc::now().to_rfc3339()));

    //run
    let cmd_result = Command::new("sh")
        .arg("-c")
        .arg(format!("gtime -f'%M' timeout --preserve-status {} {} {} {}", timeout, command, output_config.canonicalize().unwrap().to_str().unwrap(), output_path.to_str().unwrap()))
        .output()
        .expect("failed to execute process");

    let stderr = String::from_utf8_lossy(&cmd_result.stderr);

    let re = Regex::new(r"(\d+)\s*$").expect("Invalid regex");
    let memory = re.captures(&stderr).and_then(|c| c.get(1)).map(|g| g.as_str().parse::<usize>().unwrap()).unwrap_or(0);

    let end = Utc::now().to_rfc3339();

    let file = File::open(output_path);
    let output = file.ok().and_then(|f| serde_json::from_reader(f).ok());

    let file = File::open(output_config);
    let input = file.ok().and_then(|f| serde_yaml::from_reader(f).ok()).unwrap();


    ExecutionResult {
        start,
        end,
        algorithm: command.to_string(),
        input,
        output,
        executionStatus: match cmd_result.status.code() {
            Some(0) => ExecutionStatus::Success,
            Some(1) => ExecutionStatus::MemoryLimit,
            Some(143) => ExecutionStatus::RuntimeLimit,
            Some(a) => ExecutionStatus::Other(a as usize),
            None => ExecutionStatus::Unknown,
        },
        memory,
        stdout: String::from_utf8_lossy(&cmd_result.stdout).parse().unwrap(),
        stderr: stderr.to_string()
    }
}

fn main() {
    if let Err(err) = run() {
        println!("{}", err);
        process::exit(1);
    }
}
