use std::error::Error;
use std::path::{Path};
use csv::ByteRecord;
use rand::thread_rng;
use rand::seq::SliceRandom;


use plistore::CSVConfig;

pub trait DatasetSampler {
    fn sample(&self, input: &Path, output: &Path, csv_config: &CSVConfig) -> Result<(), Box<dyn Error>>;
}

pub struct ColumnSampler {
    column_limit: usize,
}

impl ColumnSampler {
    pub fn new(column_limit: usize) -> ColumnSampler {
        ColumnSampler { column_limit }
    }
}

impl DatasetSampler for ColumnSampler {
    fn sample(&self, input: &Path, output: &Path, csv_config: &CSVConfig) -> Result<(), Box<dyn Error>> {
        let mut writer = csv_config.get_writer(output)?;
        for mut i in csv_config.get_reader(input).unwrap().into_byte_records().flatten() {
            i.truncate(self.column_limit);

            writer.write_byte_record(&i)?;
        }
        Ok(())
    }
}

pub struct RandomColumnSampler {
    column_limit: usize,
}

impl RandomColumnSampler {
    pub fn new(column_limit: usize) -> RandomColumnSampler {
        RandomColumnSampler { column_limit }
    }
}

impl DatasetSampler for RandomColumnSampler {
    fn sample(&self, input: &Path, output: &Path, csv_config: &CSVConfig) -> Result<(), Box<dyn Error>> {
        let column_count = csv_config.get_reader(input).unwrap().into_byte_records().flatten().next().unwrap().len();
        let mut rng = thread_rng();
        let mut indices: Vec<usize> = Vec::from_iter(0..column_count);
        indices.shuffle(&mut rng);
        indices.truncate(self.column_limit);

        let mut writer = csv_config.get_writer(output)?;
        for i in csv_config.get_reader(input).unwrap().into_byte_records().flatten() {
            let array : Vec<&[u8]> = indices.iter().map(|idx| i.get(*idx).unwrap()).collect();
            writer.write_byte_record(&ByteRecord::from(array))?;
        }
        Ok(())
    }
}

pub struct RowSampler {
    row_limit: usize
}

impl RowSampler {
    pub fn new(row_limit: usize) -> RowSampler {
        RowSampler { row_limit }
    }
}

impl DatasetSampler for RowSampler {
    fn sample(&self, input: &Path, output: &Path, csv_config: &CSVConfig) -> Result<(), Box<dyn Error>> {
        let mut writer = csv_config.get_writer(output)?;
        let mut count = 0;
        for i in csv_config.get_reader(input)?.into_byte_records() {
            writer.write_byte_record(&i?)?;
            count += 1;
            if count == self.row_limit {
                break;
            }
        }
        Ok(())
    }
}
